/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.sql.*;
import java.util.*;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xquark.extractor.common.*;
import org.xquark.extractor.metadata.*;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.util.jdbc.DBMSInfoMap;
import org.xquark.util.jdbc.JDBCProperties;
import org.xquark.xml.xdbc.*;

/**
 * Extractor XML/DBC DataSource implementation.
 */
public class Extractor implements XMLDataSource, Constants {
	public static final String DATASOURCE_XSD = "/org/xquark/extractor/resources/Datasource.xsd";

	private static final String RCSRevision = "$Revision: 1.14 $";
	private static final String RCSName = "$Name:  $";

	/** List of supported JDBC drivers */

	private static final String[] DRIVER_LIST = { "com.mysql.jdbc.Driver", "com.microsoft.jdbc.sqlserver.SQLServerDriver", "com.sybase.jdbc2.jdbc.SybDriver", "oracle.jdbc.driver.OracleDriver" };

	static {
		MessageLibrary.setup("org.xquark.extractor.common.resources.MessageLibrary", new Locale("en", "US"));
	}

	public static boolean USE_SCROLLABLE_RESULTSET = false;

	// QUESTION : If the is static data, shouldn't accessors be synchronized ?
	static HashMap _configMap = new HashMap();
	static HashMap _propsMap = new HashMap();
	static Properties _defaultProperties = new Properties();
	static {
		/* set default properies*/
		_defaultProperties.setProperty(LABEL_SPEC_NAME, "WrapperSQL1");
		_defaultProperties.put(LABEL_SPEC_SUBST_NAMECASE, new Integer(CASE_MIXED));
		Properties substitutions = new Properties();
		substitutions.setProperty("$", "_");
		substitutions.setProperty("#", "_");
		_defaultProperties.put(LABEL_SPEC_SUBST, substitutions);
		//_defaultProperties.setProperty(LABEL_SPEC_ALWAYS_SCHEMA,"no");
		//_defaultProperties.setProperty(LABEL_SPEC_NO_NAMESPACE,"true");
		//_defaultProperties.setProperty(LABEL_SPEC_SCHEMA, "USE-CASES");
	}

	private static SchemaManager loadingManager = null;
	static {
		try {
			SchemaValidationContext context = new SchemaValidationContext();
			loadingManager = context.getSchemaManager();

			URL url = Extractor.class.getResource(DATASOURCE_XSD);
			InputSource source = new InputSource(url.toString());
			loadingManager.loadSchema(context, source);
		} catch (SAXException saxe) {
			//throw new MetaDataException (saxe.getMessage ()) ;
		}
	}

	protected String name = null;
	private Configuration _currentConfig;
	private Map _viewDefinitions = null;
	private XMLDBCUri _uri = null;
	private Connection _jdbcConnection = null;
	private ClassLoader classLoader = null;
	/** List available drivers loaded through the classLoader */
	private List drivers = null;

	public Extractor(XMLDBCUri uri) throws XMLDBCException {
		_uri = uri;
		loadDrivers(null);
	}

	public Extractor(Connection jdbcConnection) throws XMLDBCException {
		this(jdbcConnection, null, null);
	}

	public Extractor(Connection jdbcConnection, String configFile) throws XMLDBCException {
		this(jdbcConnection, configFile, null);
	}

	public Extractor(Connection jdbcConnection, String configFile, ClassLoader classLoader) throws XMLDBCException {
		_jdbcConnection = jdbcConnection;

		loadDrivers(classLoader);

		if (configFile != null) {
			_uri = new XMLDBCUri(ExtractorReadOnlyConnection.URL_PREFIX + configFile);
		} else if (jdbcConnection != null) {
			try {
				DatabaseMetaData dbm = jdbcConnection.getMetaData();
				_uri = new XMLDBCUri(ExtractorReadOnlyConnection.URL_PREFIX + dbm.getURL());
			} catch (SQLException ex) {
				throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
			}
		} else {
			throw new NullPointerException("At least one of jdbcConnection and configFile must be not null");
		}
	}

	private void loadDrivers(ClassLoader classLoader) {
		this.classLoader = classLoader != null ? classLoader : this.getClass().getClassLoader();
		drivers = new ArrayList();
		for (int i = 0; i < DRIVER_LIST.length; i++) {
			try {
				Class c = Class.forName(DRIVER_LIST[i], true, this.classLoader);
				drivers.add(c.newInstance());
			} catch (ClassNotFoundException e) {
			} catch (IllegalAccessException e) {
			} catch (InstantiationException e) {
			}
		}
	}

	/**
	 * @param writer
	 */
	public void setLogWriter(PrintWriter writer) {
	}

	/**
	 * @return java.io.PrintWriter
	 */
	public PrintWriter getLogWriter() {
		return null;
	}

	/**
	 * @param seconds
	 */
	public void setLoginTimeout(int seconds) {
	}

	/**
	 * @return int
	 */
	public int getLoginTimeout() {
		return -1;
	}

	public void setName(String name) {
		this.name = name;
	}

	/**
	 * Must only be called by SQLWrapperHandler
	 */
	public void setSelection(Selection selection) {
		_currentConfig.selection = selection;
	}

	static public Properties getProperties() {
		return _defaultProperties;
	}

	public ExtractorConfigurationFileHandler getHandler() {
		return new ExtractorConfigurationFileHandler(this);
	}

	public XMLConnection getConnection(String user, String password) throws XMLDBCException {
		Trace.enter(this, "getConnection(String user, String password)", user);
		Log.message(MessageLibrary.getMessage("S_CONN_URL", _uri.getUriString()));

		XMLConnection retVal = null;
		Connection jdbcConnection = null;
		Properties properties = null;
		String configFileName = null;
		// Load configuration file
		if (_uri.containsConfigFileName()) {
			configFileName = _uri.getConfigFileName();
			properties = (Properties) _propsMap.get(configFileName);
			if (properties == null) {
				properties = parseConfigFile(configFileName);
				_propsMap.put(configFileName, properties);
			}
		}
		try {
			if (_jdbcConnection != null) {
				// Use user-provided connection if present
				jdbcConnection = _jdbcConnection;
			} else if (_uri.containsJDBCUri()) {
				// Use JDBC URI + user + password
				if (user == null || password == null) {
					throw new XMLDBCException("please indicate username and password.");
				}
				jdbcConnection = createJdbcConnection(null, _uri.getJdbcUri(), user, password);
			} else if (properties != null) {
				// Use configuration file info, overrides user and password if given
				if (user != null && password != null) {
					properties = (Properties) properties.clone();
					properties.setProperty(LABEL_SPEC_USER, user);
					properties.setProperty(LABEL_SPEC_PASSWORD, password);
				}
				jdbcConnection = createJdbcConnection(properties.getProperty(LABEL_SPEC_DRIVER), properties.getProperty(LABEL_SPEC_URL), properties.getProperty(LABEL_SPEC_USER), properties.getProperty(LABEL_SPEC_PASSWORD));
			} else {
				throw new XMLDBCException("Missing connection information");
			}
			// Retrieves metadata if already loaded
			DatabaseMetaData dbm = jdbcConnection.getMetaData();
			String jdbcUri = dbm.getURL();
			user = dbm.getUserName();
			String key = jdbcUri + "|" + user;
			if (properties != null) {
				key += "|" + configFileName;
			}
			_currentConfig = (Configuration) _configMap.get(key);
			if (_currentConfig == null) {
				// load metadata if not already loaded
				_currentConfig = new Configuration();
				_currentConfig.viewDefinitions = _viewDefinitions;
				_currentConfig.uri = _uri;
				if (properties == null) {
					properties = (Properties) _defaultProperties.clone();
					properties.setProperty(LABEL_SPEC_URL, jdbcUri);
					properties.setProperty(LABEL_SPEC_USER, user);
				}
				_currentConfig.properties = properties;
				_currentConfig.selection = (Selection) properties.get("selections");
				checkJdbcConnection(dbm, properties);
				_currentConfig.metaDataManager = createMetaDataManager(jdbcConnection, _currentConfig);
				_currentConfig.metaDataForParser = _currentConfig.metaDataManager.buildMetaDataForParser(_currentConfig.properties.getProperty(LABEL_SPEC_NAME));
				_configMap.put(key, _currentConfig);
			}
			retVal = new ExtractorReadOnlyConnection(jdbcConnection, _currentConfig, _jdbcConnection != null);
			_currentConfig.metaDataForParser.setConnection(retVal);
		} catch (SQLException ex) {
			throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
		}
		Trace.exit(this, "getConnection(String user, String password)", user);
		return retVal;
	}

	/**
	   * @return org.xquark.xml.xdbc.XMLConnection
	   * @throws org.xquark.xml.xdbc.XMLDBCException
	   */
	public XMLConnection getConnection() throws XMLDBCException {
		Trace.enter(this, "getConnection()");
		Log.message(MessageLibrary.getMessage("S_CONN_URL", _uri.getUriString()));
		XMLConnection retVal = getConnection(null, null);
		Trace.exit(this, "getConnection()");
		return retVal;
	}

	/**
	 * Give an XML config file to the wrapper.
	 *
	 * @param xml the XML file used for configuration.
	 */
	public final Properties parseConfigFile(String file) throws XMLDBCException {
		Trace.enter(this, "parseConfigFile(String file)", file);
		Properties xmlproperties = new Properties();
		try {
			XMLReader reader = null;
			// Create a JAXP SAXParserFactory and configure it
			//TODO should be static
			SAXParserFactory spf = SAXParserFactory.newInstance();
			spf.setNamespaceAware(true);

			// Create a JAXP SAXParser
			SAXParser saxParser = spf.newSAXParser();

			// Get the encapsulated SAX XMLReader
			reader = saxParser.getXMLReader();

			//ExtractorHandler handler = getHandler() ;
			ExtractorConfigurationFileHandler handler = getHandler();

			ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, new SchemaValidationContext(loadingManager));
			filter.setContentHandler(handler);
			filter.setErrorHandler(handler);

			filter.parse(file);
			_viewDefinitions = handler.getViewDefinitions();
			xmlproperties = handler.getProperties();
		} catch (SAXParseException ex) {
			String msg = "Error in parsing file " + ex.getSystemId() + " at line " + Integer.toString(ex.getLineNumber()) +
				//" column " + Integer.toString(ex.getColumnNumber()) +
	" : " + ex.getMessage();
			throw new XMLDBCException(msg, ex);
		} catch (SAXException ex) {
			//            e.printStackTrace();
			throw new XMLDBCException("config failed", ex);
		} catch (IOException ex) {
			//            e.printStackTrace();
			throw new XMLDBCException("config failed", ex);
		} catch (ParserConfigurationException ex) {
			//            e.printStackTrace();
			throw new XMLDBCException("config failed", ex);
		}

		Trace.exit(this, "parseConfigFile(String file)", file);
		return xmlproperties;
	}

	/**
	 * Set the DATABASE properties basing on JDBC DatabaseMetadata.
	 * @param dbm
	 * @param properties
	 * @throws XMLDBCException
	 */
	static void checkJdbcConnection(DatabaseMetaData dbm, Properties properties) throws XMLDBCException {
		Trace.enter("Extractor", "checkJdbcConnection(DatabaseMetaData jdbcMetaData, Properties properties)");
		try {
			/* database */
			String dbName = dbm.getDatabaseProductName();
			String dbVersion = dbm.getDatabaseProductVersion();

			properties.setProperty(LABEL_SPEC_DATABASE_NAME, dbName);
			properties.setProperty(LABEL_SPEC_DATABASE_VERSION, dbVersion);

			Log.message(MessageLibrary.getMessage("D_DATABASE", dbName, dbVersion));
			/* driver */
			String driverName = dbm.getDriverName();
			String driverVersion = dbm.getDriverVersion();
			int majorVersion = dbm.getDriverMajorVersion();
			int minorVersion = dbm.getDriverMinorVersion();
			Log.message(MessageLibrary.getMessage("D_DRIVER", driverName, driverVersion));
		} catch (SQLException ex) {
			throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
		}

		Trace.exit("Extractor", "checkJdbcConnection(DatabaseMetaData jdbcMetaData, Properties properties)");
	}

	private Connection createJdbcConnection(String driver, String connectionString, String user, String password) throws XMLDBCException {
		Trace.enter(this, "createJdbcConnection (String driver, String connectionString, String user,String password)");

		Connection retVal = null;
		Properties jdbcProps = new Properties();
		jdbcProps.put("user", user);
		jdbcProps.put("password", password);
		if (null != driver) {
			try {
				Class c = Class.forName(driver, true, this.classLoader);
				Driver d = (Driver) c.newInstance();
				if (d.acceptsURL(connectionString)) {
					retVal = d.connect(connectionString, jdbcProps);
				}
			} catch (SQLException ex) {
				throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
			} catch (ClassNotFoundException ex) {
				throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
			} catch (IllegalAccessException ex) {
				throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
			} catch (InstantiationException ex) {
				throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
			}
		} else {
			try {
				for (int i = 0; retVal == null && i < drivers.size(); i++) {
					Driver d = (Driver) drivers.get(i);
					if (d.acceptsURL(connectionString)) {
						retVal = d.connect(connectionString, jdbcProps);
					}
				}
			} catch (SQLException ex) {
				throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()), ex);
			}
		}
		if (retVal == null) {
			throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL", "No suitable driver found"));
		}

		Trace.exit(this, "createJdbcConnection (String driver, String connectionString, String user,String password)");
		return retVal;
	}

	public static MetaDataManager createMetaDataManager(Connection jdbcConnection, Configuration config) throws XMLDBCException {
		Trace.enter("Extractor", "CreateMetaDataManager(DatabaseMetaData databaseMetaData, Configuration config)");

		MetaDataManager retVal = null;

		/* create a CustomizedEncoder*/
		CustomizedEncoder encoder = new CustomizedEncoder();
		Properties substutionMap = (Properties) config.properties.get(LABEL_SPEC_SUBST);
		encoder.setSubstitutionMap(substutionMap);
		// STOPPED HERE: move all constants (config) to Constants and use implements 
		Integer nameCase = (Integer) config.properties.get(LABEL_SPEC_SUBST_NAMECASE);
		if (null == nameCase) {
			nameCase = new Integer(CASE_MIXED);
		}
		encoder.setNameCase(nameCase.intValue());

		/* Select and call the DBMS-specific metadata manager */
		String dbName = config.properties.getProperty(LABEL_SPEC_DATABASE_NAME);
		if (dbName.startsWith("Oracle")) {
			retVal = createMetaDataManagerForOracle(jdbcConnection, config, encoder);
		} else if (dbName.startsWith("Microsoft SQL Server")) {
			retVal = createMetaDataManagerForMicrosoftSqlServer(jdbcConnection, config, encoder);
		} else if (dbName.startsWith("Adaptive Server Enterprise")) {
			retVal = createMetaDataManagerForSybase(jdbcConnection, config, encoder);
		} else if (dbName.startsWith("PROGRESS")) {
			retVal = createMetaDataManagerForProgress(jdbcConnection, config, encoder);
		} else if (dbName.startsWith("MySQL")) {
			retVal = createMetaDataManagerForMySql(jdbcConnection, config, encoder);
		} else {
			throw new XMLDBCNotSupportedException(MessageLibrary.getMessage("D_N_SUP_DB_VENDOR", dbName), null);
		}

		Trace.exit("Extractor", "CreateMetaDataManager(DatabaseMetaData databaseMetaData, Configuration config)");
		return retVal;
	}

	private static MetaDataManager createMetaDataManagerForMicrosoftSqlServer(Connection jdbcConnection, Configuration config, QNameEncoder encoder) {
		Trace.enter("Extractor", "MetaDataManagerForMicrosoftSqlServer(DatabaseMetaData databaseMetaData, Configuration config)");

		MetaDataManager retVal;
		try {
			retVal = new MetaDataManager(DBMSInfoMap.getDBMSInfo(jdbcConnection));
		} catch (SQLException ex) {
			throw new MetaDataException("JDBC error.", ex);
		}

		/* If no selection in the configuration file, include the whole current
		 * schema.
		 */
		if (config.selection == null) {
			try {
				config.selection = buildDefaultSelection(jdbcConnection.getCatalog(), getSQLServerSchema(jdbcConnection));
			} catch (SQLException ex) {
				throw new MetaDataException(ex.getMessage(), ex);
			}
		}

		/* Load metadata in MetadataManager */
		try {
			Loader metadataLoader = new org.xquark.extractor.microsoft.Loader(jdbcConnection, retVal.getDbmsInfo(), encoder);
			retVal.setMetaData(metadataLoader.loadSite(config.selection, new JDBCProperties() {
				public boolean useResultSet4ColumnMetadata() {
					return false;
				}
				public boolean useStaticMappingFromNativeType() {
					return true;
				}
			}));
			// retVal.setTableMap(metadataLoader._tableMap);	// QUESTION : why removed ?
			retVal.setViewMap(config.viewDefinitions);
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}
		Trace.exit("Extractor", "MetaDataManagerForMicrosoftSqlServer(DatabaseMetaData databaseMetaData, Configuration config)");
		return retVal;
	}

	private static MetaDataManager createMetaDataManagerForSybase(Connection jdbcConnection, Configuration config, QNameEncoder encoder) {
		Trace.enter("Extractor", "createMetaDataManagerForSybase(DatabaseMetaData databaseMetaData, Configuration config)");

		MetaDataManager retVal;
		try {
			retVal = new MetaDataManager(DBMSInfoMap.getDBMSInfo(jdbcConnection));
		} catch (SQLException ex) {
			throw new MetaDataException("JDBC error.", ex);
		}

		Loader metadataLoader = null;

		/* If no selection in the configuration file, include the whole current
		 * schema.
		 */
		if (config.selection == null) {
			try {
				config.selection = buildDefaultSelection(jdbcConnection.getCatalog(), getSQLServerSchema(jdbcConnection));
			} catch (SQLException ex) {
				throw new MetaDataException(ex.getMessage(), ex);
			}
		}

		/* Load metadata in MetadataManager */
		try {
			metadataLoader = new org.xquark.extractor.sybase.Loader(jdbcConnection, retVal.getDbmsInfo(), encoder);
			retVal.setMetaData(metadataLoader.loadSite(config.selection, new JDBCProperties() {
				public boolean useResultSet4ColumnMetadata() {
					return false;
				}
				public boolean useStaticMappingFromNativeType() {
					return true;
				}
			}));
			// retVal.setTableMap(metadataLoader._tableMap);	// QUESTION : why removed ?
			retVal.setViewMap(config.viewDefinitions);
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}
		Trace.exit("Extractor", "MetaDataManagerForSybase(DatabaseMetaData databaseMetaData, Configuration config)");
		return retVal;
	}

	// NOTE: Sybase ASE and MS SQLServer are very similar since "Bill" bought Sybase technology once upon a time...
	private static String getSQLServerSchema(Connection jdbcConnection) {
		String ret = null;
		Statement statement = null;
		try {
			statement = jdbcConnection.createStatement();
			ResultSet rs = statement.executeQuery("select user");
			rs.next();
			ret = rs.getString(1);
		} catch (SQLException e) {
			throw new MetaDataException("Could not get current schema name.", e);
		} finally {
			if (statement != null)
				try {
					statement.close();
				} catch (SQLException e1) {
				}
		}
		return ret;
	}

	private static MetaDataManager createMetaDataManagerForOracle(Connection jdbcConnection, Configuration config, QNameEncoder encoder) {

		Trace.enter("Extractor", "createMetaDataManagerForOracle(DatabaseMetaData databaseMetaData, Configuration config)");

		MetaDataManager retVal;
		try {
			retVal = new MetaDataManager(DBMSInfoMap.getDBMSInfo(jdbcConnection));
		} catch (SQLException ex) {
			throw new MetaDataException("JDBC error.", ex);
		}
		Loader metadataLoader = null;

		/* If no selection in the configuration file, include the whole current
		 * schema.
		 */
		if (config.selection == null)
			config.selection = buildDefaultSelection(null, config.properties.getProperty(LABEL_SPEC_USER));

		/* Load metadata in MetadataManager */
		try {
			metadataLoader = new org.xquark.extractor.oracle.Loader(jdbcConnection, retVal.getDbmsInfo(), encoder);
			retVal.setMetaData(metadataLoader.loadSite(config.selection, new JDBCProperties() {
				public boolean useResultSet4ColumnMetadata() {
					return false;
				}
				public boolean useStaticMappingFromNativeType() {
					return true;
				}
			}));
			retVal.setTableMap(metadataLoader._tableMap);
			retVal.setViewMap(config.viewDefinitions);
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		Trace.exit("Extractor", "createMetaDataManagerForOracle(DatabaseMetaData databaseMetaData, Configuration config)");
		return retVal;
	}

	private static MetaDataManager createMetaDataManagerForProgress(Connection jdbcConnection, Configuration config, QNameEncoder encoder) {

		Trace.enter("Extractor", "createMetaDataManagerForOracle(DatabaseMetaData databaseMetaData, Configuration config)");

		MetaDataManager retVal;
		try {
			retVal = new MetaDataManager(DBMSInfoMap.getDBMSInfo(jdbcConnection));
		} catch (SQLException ex) {
			throw new MetaDataException("JDBC error.", ex);
		}

		Loader metadataLoader = null;

		/* If no selection in the configuration file, include the whole current
		 * schema.
		 */
		if (config.selection == null)
			config.selection = buildDefaultSelection(null, config.properties.getProperty(LABEL_SPEC_USER));

		/* Load metadata in MetadataManager */
		try {
			metadataLoader = new org.xquark.extractor.progress.Loader(jdbcConnection, retVal.getDbmsInfo(), encoder);
			retVal.setMetaData(metadataLoader.loadSite(config.selection, new JDBCProperties() {
				public boolean useResultSet4ColumnMetadata() {
					return false;
				}
				public boolean useStaticMappingFromNativeType() {
					return true;
				}
			}));
			retVal.setTableMap(metadataLoader._tableMap);
			retVal.setViewMap(config.viewDefinitions);
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		Trace.exit("Extractor", "createMetaDataManagerForOracle(DatabaseMetaData databaseMetaData, Configuration config)");
		return retVal;
	}

	private static MetaDataManager createMetaDataManagerForMySql(Connection jdbcConnection, Configuration config, QNameEncoder encoder) {

		Trace.enter("Extractor", "createMetaDataManagerForMySql(DatabaseMetaData databaseMetaData, Configuration config)");

		MetaDataManager retVal;
		try {
			retVal = new MetaDataManager(DBMSInfoMap.getDBMSInfo(jdbcConnection));
		} catch (SQLException ex) {
			throw new MetaDataException("JDBC error.", ex);
		}

		Loader metadataLoader = null;

		/* If no selection in the configuration file, include the whole current
		 * catalog.
		 */
		if (config.selection == null) {
			try {
					config.selection = buildDefaultSelection(jdbcConnection.getCatalog(), null // No schema in MySQL
	);
			} catch (SQLException ex) {
				throw new MetaDataException(ex.getMessage(), ex);
			}
		}

		/* Load metadata in MetadataManager */
		try {
			metadataLoader = new org.xquark.extractor.mysql.Loader(jdbcConnection, retVal.getDbmsInfo(), encoder);
			retVal.setMetaData(metadataLoader.loadSite(config.selection, new JDBCProperties() {
				public boolean useResultSet4ColumnMetadata() {
					return false;
				}
				public boolean useStaticMappingFromNativeType() {
					return true;
				}
			}));
			retVal.setTableMap(metadataLoader._tableMap);
			retVal.setViewMap(config.viewDefinitions);
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		Trace.exit("Extractor", "createMetaDataManagerForMySql(DatabaseMetaData databaseMetaData, Configuration config)");
		return retVal;
	}

	private static Selection buildDefaultSelection(String catalog, String schema) {
		Selection site = new Selection(null, null, Selection.INCLUDES, new HashMap());
		Selection catalogSel = new Selection(catalog, null, Selection.INCLUDES, new HashMap());
		site.addToMap(catalog, catalogSel);

		Selection schemaSel = new Selection(schema, null, Selection.INCLUDES, null);
		catalogSel.addToMap(schema, schemaSel);
		return site;
	}

	/**
	 * Show the configuration of the wrapper
	 */
	public void display() {
		if (_currentConfig == null)
			return;
		Properties properties = _currentConfig.properties;
		if (properties == null)
			return;
		for (Enumeration e = properties.propertyNames(); e.hasMoreElements();) {
			String key = (String) e.nextElement();
			System.out.println("[" + key + "] : [" + properties.getProperty(key) + "]");
		}
	}
}
