/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator;

import java.io.PrintStream;
import java.io.StringReader;
import java.util.ArrayList;

import org.xquark.mediator.algebra.AlgebraManager;
import org.xquark.mediator.decomposer.Utils;
import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.MetaDataImpl;
import org.xquark.xquery.metadata.VarCounter;
import org.xquark.xquery.metadata.resolver.MetadataAccess;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.util.TransformXQueryExpression;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;
import org.xquark.xquery.xdbc.XResultSetImpl;

/**
 * This class implements XMLStatement interface that is used for executing
 * constant XQuery or XPath statements
 * and obtaining the results produced (update statements are not
 * supported yet).<BR>
 *
 */
public class MediatorStatement implements XMLStatement {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.21 $";
	private static final String RCSName = "$Name:  $";
	
	// **********************************************************************
	// * CLASS VARIABLE
	// **********************************************************************

	private MediatorConnection connection = null ;
	private boolean qaOnly = false;

	//public XDBCResultSetInterface debug_resultset = null ;
	//public XQueryExpression debug_qmem = null ;
	private String baseURI = null;

	// ************************************************************************
	// * INITIALISATION
	// ************************************************************************
	public MediatorStatement(MediatorConnection connection) throws XMLDBCException {
		this.connection = connection ;
	}

	// ************************************************************************
	// * XMLStatement IMPLEMENTATION
	// ************************************************************************

	/*********************************************************************/
	/*************************** Queries execution management ************/
	/*********************************************************************/
	/**
	 * Executes a query statement of the type specified when this
	 * object was built with the createStatement method (see XMLConnection interface)
	 * that may return results.
	 * If the query generates results, use the <I>getResultSet()</I> or
	 * <I>getDocumentSet()</I> methods to get the set of results.
	 * @param query the query string.
	 * @return true if there are available resuts, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public boolean execute(String query) throws XMLDBCException {
		return execute(query, XMLConnection.XQUERY_STRING_TYPE) ;
	}

	/**
	 * Executes a query statement of the specified type
	 * that may return results.
	 * If the query generates results, use the <I>getResultSet()</I> or
	 * <I>getDocumentSet()</I> methods to get the set of results.
	 * @param query the query string.
	 * @param queryType the type of query for this statement (see constants in XMLConnection interface).
	 * @return true if there are available resuts, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public boolean execute(String query, int queryType) throws XMLDBCException {
		throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	/**
	 * Executes a query statement of the type specified when this
	 * object was built with the createStatement method (see XMLConnection interface)
	 * that returns results.
	 * @param query the query string
	 * @return an XMLResultSet object that contains the data produced by the given query.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLResultSet executeQuery(String query) throws XMLDBCException {
		return executeQuery(query, XMLConnection.XQUERY_STRING_TYPE) ;
	}
	// for regression purposes -> printing the plan
	private PrintStream plnout = null;
	public XMLResultSet executeQuery(String query, PrintStream plnout) throws XMLDBCException {
		this.plnout = plnout;
		return executeQuery(query, XMLConnection.XQUERY_STRING_TYPE) ;
	}

	/**
	 * Executes a query statement of the specified type
	 * that returns results.
	 * If the query generates results, use the <I>getResultSet()</I> or
	 * <I>getDocumentSet()</I> methods to get the set of results.
	 * @param query the query string.
	 * @param queryType the type of query for this statement (see constants in XMLConnection interface).
	 * @return true if there are available resuts, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLResultSet executeQuery(String query, int queryType) throws XMLDBCException {
		org.xquark.util.misc.Debug.print("MediatorStatement.executeQuery (" + query + ")") ;

		try {
			xqueryParser ql = new xqueryParser(new StringReader(query));
			ql.setBaseURI(baseURI);
			ql.setFactory("org.xquark.xquery.parser.ParserFactory") ;
			MetaDataImpl metadata = (MetaDataImpl) connection.getMetaData() ;
			
			//org.xquark.xml.misc.Debug.print("Metadata = " + metadata.getMetaData().getAsString());
			//TEST ql.setRenameVars(false);
			XQueryModule module = ql.Start((MetadataAccess)metadata, metadata.getSchemaManager(), new VarCounter() ,qaOnly);
			ArrayList expressions = module.getExpressions();
			if (expressions == null || expressions.isEmpty())
				return null;
			TypeVisitor typevisitor = ql.getTypeVisitor();
			
			org.xquark.util.misc.Debug.print("=-=-=-=-= request =-=-=-=-= ") ;
			org.xquark.util.misc.Debug.print("Request = \"" + module + "\"");
			// get declarations
			
			// if only one source is concerned, no need to bother about mediating, just pipeline requests to it.
			// but first normalize and restructure
			//----------------------------------------------------------
			// START --> FOR SINGLE SOURCE VERIFICATION
			//----------------------------------------------------------
			// comment this bloc if you want the mediator not to overpass single source optimization (for debugging on a simple example)
			
			module.normalize(typevisitor,connection.getSourceName());
			// TODO handle several expressions with union !!!
			if (expressions.size() > 1) {
				throw new XMLDBCException("Fusion only handles modules with a single query (for now...)");
			}
			for (int j = 0; j < expressions.size(); j++) {
				XQueryExpression exprj = (XQueryExpression) expressions.get(j) ;
				Utils.fillSources(exprj,metadata,true);
			}
			
			//GenerateSourceVisitor sourcevisitor = new GenerateSourceVisitor (unit, metadata) ;
			//request.accept (sourcevisitor) ;
			ArrayList sources = module.getSourceNames() ;
			ArrayList urls = module.getUrls() ;
			if (!qaOnly && sources != null && sources.size () == 1 && (urls == null || urls.isEmpty()))
			{
				XMLConnection con = connection.getSubConnectionByName ((String) sources.get (0)) ;
				XMLStatement stat = con.createStatement();
				stat.setBaseURI(baseURI);
				if (stat instanceof org.xquark.extractor.ExtractorStatement) return stat.executeQuery(module.toString(false,false), -1) ;
				else return  stat.executeQuery(module.toString(false,false)) ;
			}
			
			//----------------------------------------------------------
			// STOP --> FOR SINGLE SOURCE VERIFICATION
			//----------------------------------------------------------
			org.xquark.util.misc.Debug.print("=-=-=-=-= canonization =-=-=-=-= ") ;
				
				
			// !!!!!! danger, only one expression ....			
			for (int j = 0 ; j < expressions.size() ; j ++) {
				XQueryExpression exprj = (XQueryExpression) expressions.get(j) ;
				
				TransformXQueryExpression txe = exprj.transformExtractor(typevisitor) ;
				ExecutionPlan plan = new ExecutionPlan(connection, typevisitor, txe.getIdVarList(), module.getDeclarations(),baseURI,qaOnly);
				AlgebraManager depmanager = new AlgebraManager(plan,metadata, typevisitor, txe.getQMEM(),txe.getQDB(),txe.getIdVarList()) ;

				ArrayList qdbs = txe.getQDB();
				
//				org.xquark.util.misc.Debug.print("\n* qdbs : " + qdbs) ;
//				for (int si = 0; si < qdbs.size(); si++)
//					org.xquark.util.misc.Debug.print("\n*QDB["+si+"]: "+((Variable) qdbs.get(si)).getExpression());
//				org.xquark.util.misc.Debug.print("\n* QMEM : " + txe.getQMEM()) ;					

				XDBCResultSetInterface resultset = null;
				if (!qdbs.isEmpty()) {
					
//					sources.clear();
//					urls.clear();
//					for (int si = 0; si < qdbs.size(); si++) {
//						XQueryExpression tmpexpr = ((Variable) qdbs.get(si)).getExpression();
//						Utils.fillSources(tmpexpr,metadata,true);
//						ArrayList tmplist = tmpexpr.getSourceNames();
//						if (tmplist != null) 
//							for (int k=0;k<tmplist.size();k++) {
//								if (!sources.contains(tmplist.get(k)))
//									sources.add(tmplist.get(k));						
//							}
//						tmplist = tmpexpr.getUrls();
//						if (tmplist != null) 
//							for (int k=0;k<tmplist.size();k++) {
//								if (!urls.contains(tmplist.get(k)))
//									urls.add(tmplist.get(k));						
//							}
//					}
					
//					Utils.fillSources(exprj,metadata,true);
//					ArrayList sources = exprj.getSourceNames() ;
//					ArrayList urls = exprj.getUrls() ;
//					if (!qaOnly && sources != null && sources.size () == 1 && (urls == null || urls.isEmpty()))
//					{
//						XMLConnection con = connection.getSubConnectionByName ((String) sources.get (0)) ;
//						XMLStatement stat = con.createStatement();
//						return stat.executeQuery(query, XMLConnection.XQUERY_STRING_TYPE) ;
//					}


					depmanager.generate() ;
					//depmanager.generate (qdbs) ;
					depmanager.computeVarPaths();

					org.xquark.util.misc.Debug.print("=-=-=-=-= dependencies =-=-=-=-= ") ;
					org.xquark.util.misc.Debug.print( depmanager.toString()) ;
//					if (plnout != null) {
//						plnout.print(depmanager.toString());
//					}

					org.xquark.util.misc.Debug.print("=-=-=-=-= execution plan =-=-=-=-= ") ;
					depmanager.FillPlan();
					org.xquark.util.misc.Debug.print(plan.toString()) ;
					if (plnout != null) {
						plnout.print(plan.toString());
					}

					org.xquark.util.misc.Debug.print("=-=-=-=-= execution =-=-=-=-= ") ;
					if (plan.getRoot() == null) {
					    org.xquark.util.misc.Debug.print("Plan is empty, no execution possible!!!");
					    return null;
					}
					else
						plan.executeQuery() ;

					// pour les tests, voir ce que l'on reoit
					resultset = plan.getResultSet() ;
				}

				GetBoundVariablesVisitor visitor = new GetBoundVariablesVisitor(false,true) ;
				txe.getQMEM().accept(visitor) ;
				ArrayList varlist = visitor.getIDsVariables() ;
				XResultSetImpl xResultSet = new XResultSetImpl(resultset, txe.getQMEM(), varlist,this) ;

				if (true) return xResultSet ;
			}
			org.xquark.util.misc.Debug.print("\n\n\n=-=-=-=-= end of execution =-=-=-=-=");
			return null ;
		}
		catch (XQueryException e) {
			throw new XMLDBCException("MediatorStatement.executeQuery :" + e.getMessage(),e);
		}
		catch (ParseException e) {
			throw new XMLDBCException("MediatorStatement.executeQuery :" + e.getMessage(),e);
		}
	}


	/**
	 * Executes a query statement of the type specified when this
	 * object was built with the createStatement method (see XMLConnection interface)
	 * that returns complete XML documents. It is a error to call this method with a
	 * query that returns document fragments.
	 * @param query a static and read-only query in the specified type.
	 * @return an XMLResultSet object that contains the data produced by the given query.
	 * @throws XMLDBCException if a data source access error occurs, or if the returned
	 * results are not complete XML documents.
	 */
	public XMLDocumentSet executeDocumentQuery(String query) throws XMLDBCException {
		return executeDocumentQuery(query, XMLConnection.XQUERY_STRING_TYPE) ;
	}

	/**
	 * Executes a query statement of the specified type
	 * that returns complete XML documents. It is a error to call this method with a
	 * query that returns document fragments.
	 * @param query the query string.
	 * @param queryType the type of query for this statement (see constants in XMLConnection interface).
	 * @return true if there are available resuts, else false.
	 * @throws XMLDBCException if a data source access error occurs, or if the returned
	 * results are not complete XML documents.
	 */
	public XMLDocumentSet executeDocumentQuery(String query, int queryType) throws XMLDBCException {
		throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	/*********************************************************************/
	/*************************** Results management **********************/
	/*********************************************************************/

	/**
	 * Returns the current statement results as an XMLResultSet object.
	 * @return the current results as an XMLResultSet object or null if there is no result.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLResultSet getResultSet() throws XMLDBCException {
		throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	/**
	 * Returns the current statement results as an XMLDocumentSet object.
	 * It is a error to call this method for a statement that returned
	 * document fragments.
	 * @return the current results as an XMLDocumentSet object or null if there is no result.
	 * @throws XMLDBCException if a data source access error occurs, or if the returned
	 * results are not complete XML documents.
	 */
	public XMLDocumentSet getDocumentSet() throws XMLDBCException {
		throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	/*********************************************************************/
	/*************************** Statement management ********************/
	/*********************************************************************/

	/**
	 * Returns the connection object that produced this statement.
	 * @return the connection object that produced this statement.
	 */
	public XMLConnection getConnection() {
		return connection ;
	}

	/**
	 * Releases the statement XMLDBC resources immediately instead of waiting for this
	 * to happen when it is automatically closed.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void close() throws XMLDBCException {
		//throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	public void setBaseURI(String baseURI) {
		this.baseURI = baseURI;
	}
	
	public String getBaseURI() {
		return baseURI;
	}
	
	public void setQAOnly(boolean bool) {
		qaOnly = bool;
	}
}

